# -*- coding: utf-8 -*-
"""s3 class"""

__author__  = "Adrien DELLE CAVE"
__license__ = """
    Copyright (C) 2018  fjord-technologies

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA..
"""

import botocore
import logging
import pycurl
import time

from mimetools import Message

try:
    from cStringIO import StringIO
except ImportError:
    from StringIO import StringIO


LOG = logging.getLogger('xtrargon.classes.s3')

DEFAULT_HTTP_HEADERS = ('Connection: keep-alive',
                        'Content-Type: application/octet-stream')
DEFAULT_MAX_REDIRS   = 5
DEFAULT_MAX_RETRIES  = 0
DEFAULT_TIMEOUT      = 60
DEFAULT_RETRY_DELAY  = 0.0

_CONFIG_OPTIONS      = ('connect_timeout',
                        'read_timeout')


class XtrargonBotoNullEventHooks(botocore.hooks.BaseEventHooks):
    def _register(self, event_name, handler, unique_id=None, unique_id_uses_count=False):
        return

    def emit_until_response(self, event_name, **kwargs):
        return (None, None)


class XtrargonS3Config(object):
    def __init__(self, options = None, boto_config = None):
        self._boto_config    = boto_config
        self.connect_timeout = None
        self.max_redirs      = None
        self.max_retries     = None
        self.parse_headers   = None
        self.read_timeout    = None
        self.retry_delay     = None
        self.retry_factor    = None

        self.set_config(options, from_boto = True)

    def get_boto_config(self):
        return self._boto_config

    def set_config(self, options, from_boto = False):
        if not options or not isinstance(options, dict):
            options = {}

        if from_boto and self._boto_config:
            for opt in _CONFIG_OPTIONS:
                options[opt] = getattr(self._boto_config, opt)

            if self._boto_config.retries and 'max_attempts' in self._boto_config.retries:
                options['max_retries'] = self._boto_config.retries['max_attempts']

        self.parse_headers = bool(options.get('parse_headers'))

        if 'max_redirs' not in options:
            options['max_redirs'] = DEFAULT_MAX_REDIRS

        if not options.get('max_retries'):
            options['max_retries'] = DEFAULT_MAX_RETRIES

        if not options.get('retry_delay'):
            options['retry_delay'] = DEFAULT_RETRY_DELAY

        if options['retry_delay'] < 0:
            options['retry_delay'] = 0

        for x in ('connect', 'read'):
            opt = "%s_timeout" % x
            if opt in options:
                setattr(self, opt, int(options[opt]))
            else:
                setattr(self, opt, int(DEFAULT_TIMEOUT))

        self.max_redirs   = int(options['max_redirs'])
        self.max_retries  = int(options['max_retries'])
        self.retry_delay  = float(options['retry_delay'])
        self.retry_factor = bool(options.get('retry_factor'))


class XtrargonS3Command(object):
    def __init__(self, client = None, session = None, config = None, killed = False):
        if not client and not session:
            session = botocore.session.Session(include_builtin_handlers = None,
                                               event_hooks = XtrargonBotoNullEventHooks())

        if not isinstance(config, XtrargonS3Config):
            self._config = XtrargonS3Config()
        else:
            self._config = config

        if client:
            self._client = client
        else:
            self._client = session.create_client('s3',
                                                 config = self._config.get_boto_config())

        self._pc          = None
        self._raw_headers = None
        self._request     = None
        self._req_dict    = None
        self._killed      = killed

    def kill(self):
        self._killed = True

    def killed(self):
        return self._killed

    def _prepare(self, op, params = None, body = None):
        if params is None:
            params = {}

        op_model = self._client._service_model.operation_model(op)

        request_context = {
            'client_region': self._client.meta.region_name,
            'client_config': self._client.meta.config,
            'has_streaming_input': op_model.has_streaming_input,
            'auth_type': op_model.auth_type,
        }

        self._req_dict = self._client._convert_to_request_dict(
                            params,
                            op_model,
                            context=request_context)

        if body is not None:
            self._req_dict['body'] = body

        self._request = botocore.awsrequest.create_request_object(self._req_dict)
        self._client._request_signer.sign(op, self._request)

        self._pc.setopt(pycurl.URL, str(self._request.url))
        self._pc.setopt(pycurl.CONNECTTIMEOUT, self._config.connect_timeout)
        self._pc.setopt(pycurl.CUSTOMREQUEST, str(self._request.method))
        self._pc.setopt(pycurl.HEADER, 1)
        self._pc.setopt(pycurl.FOLLOWLOCATION, 1)
        self._pc.setopt(pycurl.MAXREDIRS, self._config.max_redirs)
        self._pc.setopt(pycurl.LOW_SPEED_TIME, 5)
        self._pc.setopt(pycurl.LOW_SPEED_LIMIT, 1024)
        self._pc.setopt(pycurl.NOPROGRESS, 1)
        self._pc.setopt(pycurl.NOSIGNAL, 1)
        self._pc.setopt(pycurl.TIMEOUT, self._config.read_timeout)
        self._pc.setopt(pycurl.VERBOSE, 0)
        self._pc.setopt(pycurl.WRITEFUNCTION, lambda x: None)

        if self._request.method == 'HEAD':
            self._pc.setopt(pycurl.NOBODY, 1)
        elif self._req_dict['body']:
            self._pc.setopt(pycurl.POSTFIELDS, self._req_dict['body'])

        headers = list(DEFAULT_HTTP_HEADERS)

        for x in self._request.headers.items():
            headers.append(str(": ".join(x)))

        self._pc.setopt(pycurl.HTTPHEADER, headers)

    def _parse_headers(self):
        if self._raw_headers:
            self._raw_headers.seek(0)
            self._raw_headers.readline()
            return Message(self._raw_headers)

    def _clean(self, pc = False):
        if self._raw_headers:
            self._raw_headers.close()

        if pc and self._pc:
            self._pc.close()

    def run(self, op, params = None, body = None):
        r                 = {'headers':    None,
                             'code':       None,
                             'nb_retries': 0}
        if self._killed:
            return r

        error             = {'Error': {'Code': None,
                                       'Message': ''},
                             'ResponseMetadata': {}}

        self._pc          = None
        self._raw_headers = None
        self._request     = None
        self._req_dict    = None
        retry_delay       = self._config.retry_delay

        try:
            self._pc = pycurl.Curl()
            self._prepare(op, params, body)

            while not self._killed:
                try:
                    if self._config.parse_headers:
                        self._raw_headers = StringIO()
                        self._pc.setopt(pycurl.HEADERFUNCTION, self._raw_headers.write)

                    self._pc.perform()
                    r['code'] = self._pc.getinfo(self._pc.HTTP_CODE)

                    if self._config.parse_headers:
                        r['headers'] = self._parse_headers()

                    if r['code'] < 400:
                        break
                    error['Error']['Code'] = str(r['code'])
                    LOG.error("return code: %r", r['code'])
                    raise botocore.exceptions.ClientError(error, op)
                except (KeyboardInterrupt, SystemExit):
                    raise
                except Exception, e:
                    LOG.warning("%r", e)
                    r['nb_retries'] += 1
                    self._clean()
                    if r['nb_retries'] >= self._config.max_retries:
                        raise
                    if self._config.retry_factor:
                        retry_delay = (self._config.retry_delay
                                       * (2 * (r['nb_retries'] - 1)))
                    if retry_delay:
                        time.sleep(retry_delay)
        except pycurl.error:
            raise
        finally:
            self._request  = None
            self._req_dict = None
            self._clean(pc = True)

        return r

    def __call__(self, op, params = None, body = None):
        return self.run(op, params, body)
